import org.eclipse.emf.ecore.*
import org.eclipse.emf.ecore.util.*
import edu.tum.cs.cope.migration.*


deleteFeatureValue = {Instance instance, EStructuralFeature feature ->
	def value = instance.unset(feature)
	if(feature instanceof EReference && feature.containment) {
		if(feature.many) {
			for(v in value) {
				v.delete()
			}
		}
		else if(value != null) {
			value.delete()
		}
	}
}

isConcrete = { EClass eClass ->
	return !(eClass.'abstract' || eClass.'interface')
}


@label("New Class")
@description("In the metamodel, a new class is created. Nothing is changed in the model.")
newClassImpl = {
	@description("The package in which the class is created") EPackage ePackage,
	@description("The name of the new class") String name,
	@description("The super classes of the new class") Collection<EClass> superClasses = [],
	@description("Whether the class is abstract") boolean abstr = false ->
	
	ePackage.newEClass(name, superClasses, abstr)
}
newClass = newClassImpl


@label("New Reference")
@description("In the metamodel, a new reference is created. Nothing is changed in the model.")
newReferenceImpl = {
	@description("The class in which the reference is created") EClass eClass,
	@description("The name of the new reference") String name, 
	@description("The type of the new reference") EClass type,
	@description("The lower bound of the new reference") int lowerBound = 0,
	@description("The upper bound of the new reference") int upperBound = 1,
	@description("Whether the new reference is a containment reference") boolean containment = false,
	@description("The opposite reference of the new reference") EReference opposite = null ->
	
	eClass.newEReference(name, type, lowerBound, upperBound, containment, opposite)
}
newReference = newReferenceImpl


@label("New Attribute")
@description("In the metamodel, a new attribute is created. Nothing is changed in the model.")
newAttributeImpl = {
	@description("The class in which the attribute is created") EClass eClass,
	@description("The name of the new attribute") String name,
	@description("The type of the new attribute") EDataType type,
	@description("The lower bound of the new attribute") int lowerBound = 0,
	@description("The upper bound of the new reference") int upperBound = 1,
	@description("The default value literal") String defaultValue = null ->
	
	eClass.newEAttribute(name, type, lowerBound, upperBound, defaultValue)
}
newAttribute = newAttributeImpl


@label("Replace Feature")
@description("In the metamodel, a feature is replace by another one. In the model, the values are moved accordingly.")
replaceFeatureImpl = {
	@description("The feature to be replaced") EStructuralFeature toReplace,
	@description("The feature by which it is replaced") EStructuralFeature replaceBy ->

	def EClass subClass = toReplace.eContainingClass
	
	assert toReplace.eType == replaceBy.eType : "The features must be of the same type"
	assert toReplace.many == replaceBy.many : "The features must be of the same cardinality"
	assert subClass.eAllStructuralFeatures.contains(replaceBy) : "The feature to replace must be defined in a sub class of the one with the feature by which it is replaced"

	// metamodel adaptation
	toReplace.delete()
	
	// model migration
	for(instance in subClass.allInstances) {
		instance.set(replaceBy, instance.unset(toReplace))
	}
}
replaceFeature = replaceFeatureImpl


@label("Split String Attribute")
@description("In the metamodel, a new String-typed attribute is created. In the model, the value of another String-typed attribute is split among the two attributes by means of a regular expression.")
splitStringAttributeImpl = {
	@description("The feature whose values are split") EAttribute toSplit,
	@description("The class in which the new feature is created") EClass context,
	@description("The name of the new attribute") String attributeName,
	@description("The regular expression") String pattern ->

	// variables
	def EClass eClass = toSplit.eContainingClass

	// constraints
	assert eClass.eAllSubTypes.plus(eClass).contains(context) : "The class with the new attribute must be a subclass of the class with the attribute to be split"
	assert toSplit.eType == emf.EString : "The type of the attribute to split has to be String"

	// metamodel adaptation
	def newAttribute = context.newEAttribute(attributeName, emf.EString);
	
	// model migration
	for(instance in context.allInstances) {
		value = instance.get(toSplit);
		fragments = value.split(pattern);
		if(fragments.length > 1) {
			instance.set(toSplit, fragments[0]);
			instance.set(newEAttribute, fragments[1]);
		}
	}
}
splitStringAttribute = splitStringAttributeImpl


@description("In the metamodel, a number of features are extracted to a new class. This new class is accessible from the context class through a new containment reference. In the model, the values of the features are extracted to a new instance accordingly.")
@label("Extract Class")
extractClassImpl = {
	@description("The context class from which the features are extracted") EClass contextClass,
	@description("The features to be extracted") List<EStructuralFeature> features,
	@description("The package in which the extracted class is created") EPackage ePackage = contextClass.ePackage,
	@description("The name of the extracted class") String className,
	@description("The name of the new containment reference from context to extracted class") String referenceName ->

	// constraints
	assert contextClass.eStructuralFeatures.containsAll(features) : "The features have to belong to the same class"
	assert ePackage.getEClassifier(className) == null : "A classifier with the same name already exists"
	assert contextClass.getEStructuralFeature(referenceName) == null || features.contains(contextClass.getEStructuralFeature(referenceName)) : "A feature with the same name already exists"
	
	// metamodel adaptation
	def extractedClass = ePackage.newEClass(className)
	def reference = contextClass.newEReference(referenceName, extractedClass, 1, 1, true)
	extractedClass.eStructuralFeatures.addAll(features)
	for(feature in features) {
		if(feature instanceof EReference && feature.eOpposite != null) {
			feature.eOpposite.eType = extractedClass
		}
	}
	
	// model migration
	for(contextInstance in contextClass.allInstances) {
		def extractedInstance = extractedClass.newInstance()
		contextInstance.set(reference, extractedInstance);
		for(feature in features) {
			extractedInstance.set(feature, contextInstance.unset(feature))
		}
	}
}
extractClass = extractClassImpl


@description("In the metamodel, a class reachable through a single-valued containment reference is inlined. More specifically, its features are moved to the source class of the reference. In the model, the values of these features are moved accordingly.")
@label("Inline Class")
inlineClassImpl = {
	@description("The reference to the class to be inlined") EReference reference ->

	def EClass inlinedClass = reference.eReferenceType
	def EClass contextClass = reference.eContainingClass
	def List<EStructuralFeature> features = new ArrayList(inlinedClass.eStructuralFeatures)
	
	// contraints
	assert reference.eOpposite == null : "The reference must not have an opposite"
	assert reference.lowerBound == 1 && reference.upperBound == 1 : "The multiplicity of the reference must be single-valued and obligatory"
	assert reference.containment : "The reference must be containment"
	assert inlinedClass.eSubTypes.isEmpty() : "The class to be inlined must not have sub classes"
	assert inlinedClass.getInverse(emf.ETypedElement.eType).every{feature -> feature instanceof EAttribute || feature.eOpposite == null || features.contains(feature.eOpposite) } : "The class to be inlined must not be a type of another reference"
	
	// metamodel adaptation
	contextClass.eStructuralFeatures.addAll(features)
	for(feature in features) {
		if(feature instanceof EReference && feature.eOpposite != null) {
			feature.eOpposite.eType = contextClass
		}
	}
	reference.delete()
	inlinedClass.delete()
	
	// model migration
	for(contextElement in contextClass.allInstances) {
		inlinedElement = contextElement.unset(reference)
		if(inlinedElement != null) {
			for(feature in features) {
				contextElement.set(feature, inlinedElement.unset(feature))
			}
			inlinedElement.delete()
		}
	}
}
inlineClass = inlineClassImpl


@description("In the metamodel, an attribute that references elements by identifier is replaced by a reference. In the model, its values are replaced by references to that element.")
@label("Replace Identifier by Reference")
replaceIdentifierByReferenceImpl = {
	@description("The referencing attribute") EAttribute referencingAttribute,
	@description("The referenced attribute") EAttribute referencedAttribute ->

	// variables
	def EClass referencedClass = referencedAttribute.eContainingClass
	
	// constraints
	assert referencingAttribute.eType == referencedAttribute.eType : "Referencing and referenced attribute must be of the same type"
	
	// metamodel adaptation
	def referencingClass = referencingAttribute.eContainingClass
	def referencingReference = referencingClass.newEReference(
		referencingAttribute.name,
		referencedClass,
		referencingAttribute.lowerBound,
		referencingAttribute.upperBound
	)
	referencingAttribute.delete()
	
	// model migration
	def referencedElements = [:]
	for(referencedElement in referencedClass.allInstances) {
		referencedElements[referencedElement.get(referencedAttribute)] = referencedElement
	}
	for(referencingElement in referencingClass.allInstances) {
		def reference = referencingElement.unset(referencingAttribute)
		referencingElement.set(referencingReference, referencedElements[reference])
	}
}
replaceIdentifierByReference = replaceIdentifierByReferenceImpl


@description("In the metamodel, the composite design pattern is introduced. More specifically, a class is refined by two sub classes - one for composite and one for leaf elements, and a reference is moved to the composite class. In addition, the class is made abstract. In the model, instances of that class are migrated based on whether the reference is populated or not.")
@label("Partition Composite")
partitionCompositeImpl = {
	@description("The class which is refined") EClass eClass,
	@description("The name of the composite class") String compositeName,
	@description("The name of the leaf class") String leafName,
	@description("The reference for composite elements") EReference childReference ->
	
	// constraints
	assert eClass.eReferences.findAll{r -> r.containment}.contains(childReference) : "The child reference must be a containment reference defined by the class"
	assert eClass.eSubTypes.isEmpty() : "The class must not have sub classes"

	// metamodel adaptation
	eClass.'abstract' = true
	def ePackage = eClass.ePackage
	def compositeClass = ePackage.newEClass(compositeName, [eClass])
	compositeClass.eStructuralFeatures.add(childReference)
	def leafClass = ePackage.newEClass(leafName, [eClass])
	
	// model migration
	for(instance in eClass.instances) {
		if(instance.get(childReference).isEmpty()) {
			instance.migrate(leafClass)
		}
		else {
			instance.migrate(compositeClass)
		}
	}
}
partitionComposite = partitionCompositeImpl


@description("In the metamodel, a reference is deleted. In the model, the values of this reference are merged to a compatible reference.")
@label("Merge Reference into Another")
mergeImpl = {
	@description("The reference that is deleted") EReference toMerge,
	@description("The reference to which the values are merged") EReference mergeTo ->

	def EClass contextClass = toMerge.eContainingClass
	
	// constraints
	assert contextClass.eAllStructuralFeatures.contains(mergeTo) : "The reference to merge to must be available in the context class"
	assert mergeTo.many : "The reference to merge to must be multi-valued"
	assert toMerge.eReferenceType == mergeTo.eReferenceType || toMerge.eReferenceType.eAllSuperTypes.contains(mergeTo.eReferenceType) : "The types of the references must be compatible"
	assert toMerge != mergeTo : "The references must be different from each other"
	
	// model migration
	for(instance in contextClass.allInstances) {
		def value = instance.unset(toMerge)
		if(toMerge.many) {
			instance.get(mergeTo).addAll(value)
		}
		else {
			instance.get(mergeTo).add(value)
		}
	}
	
	// metamodel adaptation
	toMerge.delete()	
	if(toMerge.eOpposite != null) {
		toMerge.eOpposite.delete()
	}
}
merge = mergeImpl


@description("In the metamodel, a feature is moved along a single-valued reference. In the model, values are moved accordingly.")
@label("Move Feature along Reference")
moveFeatureImpl = {
	@description("The feature to be moved") EStructuralFeature feature,
	@description("The reference along which the feature is moved") EReference reference ->

	// variables
	def EClass sourceClass = feature.eContainingClass
	def EClass targetClass = reference.eReferenceType

	// constraints
	assert reference.lowerBound == 1 && reference.upperBound == 1 : "The multiplicity of the reference must be single-valued and obligatory"
	assert reference.eOpposite == null || reference.eOpposite.upperBound == 1 : "The multiplicity of its opposite reference must be single-valued"
	assert sourceClass.eAllStructuralFeatures.contains(reference) : "The reference must be available in the same class as the feature"
	assert targetClass.getEStructuralFeature(feature.name) == null : "A feature with that name already exists in the target class"
	
	// metamodel adaptation
	targetClass.eStructuralFeatures.add(feature)
	
	// model migration
	for(instance in sourceClass.allInstances) {
		def target = instance.get(reference)
		if(instance.isSet(feature)) {
			def value = instance.unset(feature)
			target.set(feature, value)
		}
	}
}
moveFeature = moveFeatureImpl


@description("In the metamodel, a feature is deleted. In the model, its values are deleted, too.")
@label("Delete Feature")
deleteFeatureImpl = {
	@description("The feature to be deleted") EStructuralFeature feature ->

	def eClass = feature.eContainingClass

	// metamodel adaptation
	feature.delete()
	
	// model migration
	for(instance in eClass.allInstances) {
		deleteFeatureValue(instance, feature)
	}
}
deleteFeature = deleteFeatureImpl


@description("In the metamodel, a number of features of a class are extracted to a new super class. In the model, nothing is changed.")
@label("Extract Super Class")
extractSuperClassImpl = {
	@description("The class from which the features are extracted") EClass subClass,
	@description("The features to be extracted") List<EStructuralFeature> toExtract,
	@description("The package in which the super class is created") EPackage ePackage = subClass.ePackage,
	@description("The name of the super class") String superClassName,
	@description("Whether the super class is abstract") boolean abstr = true,
	@description("The super classes of the sub class which become super classes of the super class") List<EClass> superSuperClasses = [] ->

	// constraints
	assert subClass.eStructuralFeatures.containsAll(toExtract) : "The features to be extracted must belong to sub class"
	assert subClass.eSuperTypes.containsAll(superSuperClasses) : "The super classes to be extracted must be a containsAll of the subclass's super types"

	// metamodel adaptation
	def superClass = ePackage.newEClass(superClassName)
	superClass.'abstract' = abstr
	superClass.eStructuralFeatures.addAll(toExtract)
	superClass.eSuperTypes.addAll(superSuperClasses)
	
	subClass.eSuperTypes.add(superClass)
	subClass.eSuperTypes.removeAll(superSuperClasses)

	for(EStructuralFeature feature : toExtract) {
		if(feature instanceof EReference) {
			EReference reference = (EReference) feature
			if(reference.eOpposite != null) {
				reference.eOpposite.eType = superClass
			}
		}
	}
}
extractSuperClass = extractSuperClassImpl


@description("In the metamodel, a number of features are replaced by features of a new super class. In the model, the values are moved to these features based on a mapping.")
@label("Use Super Class")
useSuperClassImpl = {
	@description("The class to which the super class is added") EClass subClass,
	@description("The new super class") EClass superClass,
	@description("The features to be replaced") List<EStructuralFeature> toReplace,
	@description("The features by which they are replaced (in the same order)") List<EStructuralFeature> replaceBy ->

	// constraints
	assert toReplace.size() == replaceBy.size() : "The number of features to be replaced and to replace them must be the same"
	assert subClass.eStructuralFeatures.containsAll(toReplace) : "The features to be replaced must belong to the sub class"
	assert superClass.eAllStructuralFeatures.containsAll(replaceBy) : "The features to replace must be available in the super class"
	assert toReplace.every{it.eType == replaceBy[toReplace.indexOf(it)].eType} : "The features must be of the same type"
	assert toReplace.every{it.many == replaceBy[toReplace.indexOf(it)].many} : "The features must be of the same multiplicity"

	subClass.eSuperTypes.add(superClass)
	subClass.eSuperTypes.removeAll(superClass.eAllSuperTypes)
	
	for(int i = 0; i < toReplace.size(); i++) {
		replaceFeature(toReplace[i], replaceBy[i])
	}
}
useSuperClass = useSuperClassImpl


@description("In the metamodel, a super class is inlined into its sub classes. More specifically, its features are propagated to the sub classes. In the model, the values of these features have to be adapted accordingly.")
@label("Inline Super Class")
inlineSuperClassImpl = {
	@description("The super class to be inlined") EClass superClass ->
		
	// constraints
	assert !isConcrete(superClass) : "The super class must be abstract"
	
	for(EStructuralFeature feature : new ArrayList(superClass.eStructuralFeatures)) {
		pushFeature(feature)
	}
	
	List<EClass> subClasses = superClass.getInverse(emf.EClass._eSuperTypes)
	for(EClass subClass in subClasses) {
		subClass.eSuperTypes.remove(superClass)
		for(EClass superSuperClass in superClass.eSuperTypes) {
			if(!subClass.eAllSuperTypes.contains(superSuperClass)) {
				subClass.eSuperTypes.add(superSuperClass)
			}
		}
	}
	
	superClass.delete()
}
inlineSuperClass = inlineSuperClassImpl


@description("In the metamodel, either the type or the multiplicity of a reference is generalized. In the model, nothing is changed.")
@label("Generalize Reference")
generalizeReferenceImpl = {
	@description("The reference to be generalized") EReference reference,
	@description("The new type of the reference") EClass type = reference.eType,
	@description("The new lower bound of the reference") int lowerBound = reference.lowerBound,
	@description("The new upper bound of the reference") int upperBound = reference.upperBound ->

	// constraints
	assert reference.eType.eAllSuperTypes.plus(reference.eType).contains(type) : "The type must be the same or more general"
	assert lowerBound <= reference.lowerBound && (upperBound >= reference.upperBound || upperBound == -1) : "The multiplicity must be the same or more general"

	// metamodel adaptation
	if(reference.eType != type) {
		reference.eType = type
		if(reference.eOpposite != null) {
			type.eStructuralFeatures.add(reference.eOpposite)
		}
	}
	reference.lowerBound = lowerBound
	reference.upperBound = upperBound
}
generalizeReference = generalizeReferenceImpl


@description("In the metamodel, a number of features are pulled up into a common super class. In the model, values are changed accordingly.")
@label("Pull up Feature")
pullFeatureImpl = {
	@description("The features to be pulled up") List<EStructuralFeature> features,
	@description("The super class to which the features are pulled") EClass targetClass = features[0].eContainingClass.eSuperTypes[0] ->

	def EStructuralFeature mainFeature = features[0]

	// constraints
	assert features.every{feature -> mainFeature.eType == feature.eType} : "The features' types have to be the same"
	assert features.every{feature -> mainFeature.lowerBound == feature.lowerBound && mainFeature.upperBound == feature.upperBound} : "The features' multiplicities have to be the same"
	assert mainFeature instanceof EAttribute || features.every{feature -> mainFeature.containment == feature.containment} : "The features have to be all containment references or not"
	assert mainFeature instanceof EAttribute || features.every{feature -> (feature == mainFeature || feature.eOpposite == null)} : "The features must not have opposite references"
	assert features.every{feature -> feature.eContainingClass.eSuperTypes.contains(targetClass)} : "The features' classes must have a common super type"
	
	targetClass.eStructuralFeatures.add(mainFeature)
	if(mainFeature instanceof EReference && mainFeature.eOpposite != null) {
		generalizeReference(mainFeature.eOpposite, targetClass)
	}
	for(feature in features) {
		if(feature != mainFeature) {
			replaceFeature(feature, mainFeature)
		}
	}	
}
pullFeature = pullFeatureImpl


@description("In the metamodel, a feature is pushed down to its sub classes. In the model, values are changed accordingly.")
@label("Push down Feature")
pushFeatureImpl = {
	@description("The feature to be pushed down") EStructuralFeature feature ->
	
	EClass superClass = feature.eContainingClass
	List<EClass> subClasses = superClass.getInverse(emf.EClass._eSuperTypes)
	
	// constraints
	assert feature instanceof EAttribute || feature.eOpposite == null || superClass.eSubTypes.size() == 1 : "If the feature has an opposite, then the super class may only have one sub type."
	
	def first = true
	for(subClass in subClasses) {
		if(first) {
			subClass.eStructuralFeatures.add(feature)
			if(feature instanceof EReference && feature.eOpposite != null) {
				feature.eOpposite.eType = subClass
			}
		}
		else {
			// metamodel adaptation
			def clone = feature.clone()
			subClass.eStructuralFeatures.add(clone)
	
			// model migration
			for(instance in subClass.allInstances) {
				instance.set(clone, instance.unset(feature))
			}
		}
		first = false;
	}
	
	if(isConcrete(superClass)) {
		for(instance in superClass.instances) {
			deleteFeatureValue(instance, feature)
		}
	}
}
pushFeature = pushFeatureImpl


@description("In the metamodel, inheritance from a super class is replaced by delegation to this class. More specifically, the super class is removed and a containment reference to this class is created. In the model, the contents associated to the super class are extracted to a separate instance of the super class.")
@label("Replace Inheritance by Delegation")
replaceInheritanceByDelegationImpl = {
	@description("The class from which the super class is removed") EClass subClass,
	@description("The super class to be removed") EClass superClass,
	@description("The name of the reference to the super class") String referenceName ->

	// constraints
	assert subClass.eSuperTypes.contains(superClass) : "The super class must be a super type of the sub class"
	
	// metamodel adaptation
	subClass.eSuperTypes.remove(superClass)
	def delegation = subClass.newEReference(referenceName, superClass, 1, 1, true)
	
	// model migration
	for(instance in subClass.allInstances) {
		def delegate = superClass.newInstance()
		instance.set(delegation, delegate)
		for(feature in superClass.eAllStructuralFeatures) {
			delegate.set(feature, instance.unset(feature))
		}
		for(slot in new ArrayList(instance.references)) {
			def reference = slot.eReference
			def source = slot.instance
			if(superClass == reference.eType || superClass.eAllSuperTypes.contains(reference.eType)) {
				if(reference.many) {
					source.get(reference).remove(instance)
					source.get(reference).add(delegate)
				}
				else {
					source.set(reference, delegate)
				}
				if(reference.containment) {
					instance.unset(delegation)
					instance.delete()
				}
			}
		}
	}
}
replaceInheritanceByDelegation = replaceInheritanceByDelegationImpl


// added due to GMF case study
@description("In the metamodel, a feature is moved opposite to a multi-valued reference. In the model, the values of the feature are aggregated accordingly.")
@label("Collect Feature over Reference")
collectFeatureImpl = {
	@description("The feature to be moved") EStructuralFeature feature,
	@description("The reference opposite to which the feature is moved") EReference reference ->
	
	// variables
	def EClass sourceClass = feature.eContainingClass
	def EClass targetClass = reference.eContainingClass
	
	// constraints
	assert (feature.many && reference.many) || !reference.many : "Both feature and reference must be multi-valued or the reference must be single-valued"
	assert reference.eReferenceType.eStructuralFeatures.contains(feature) : "The feature must belong to the reference's type"
	
	// metamodel adaptation
	targetClass.eStructuralFeatures.add(feature)
	
	// model migration
	for(target in targetClass.allInstances) {
		if(reference.many) {
			for(source in target.get(reference)) {
				def sourceValue = source.unset(feature)
				target.get(feature).addAll(sourceValue)
			}
		}
		else {
			def source = target.get(reference)
			if(source != null) {
				target.set(feature, source.unset(feature))
			}
		}
	}
	for(source in sourceClass.allInstances) {
		deleteFeatureValue(source, feature)
	}
}
collectFeature = collectFeatureImpl


// added 4/12/08 due to PCM case study
@description("In the metamodel, the type of an attribute is changed. In the model, the values are migrated based on EMF's default serialization.")
@label("Change Attribute Type")
changeAttributeTypeImpl = {
	@description("The attribute whose type is changed") EAttribute attribute,
	@description("The new type of the attribute") EDataType type ->

	// variables
	def EClass eClass = attribute.eContainingClass
	def EDataType oldType = attribute.eType
	
	// metamodel adaptation
	attribute.eType = type
	
	// model migration
	for(instance in eClass.allInstances) {
		if(instance.isSet(attribute)) {
			def value = instance.get(attribute)
			if(attribute.many) {
				def newValue = []
				for(v in value) {
					def stringValue = EcoreUtil.convertToString(oldType, v)
					def nv = EcoreUtil.createFromString(type, stringValue)
					newValues.add(nv)
				}
				instance.set(attribute, newValue)
			}
			else {
				def stringValue = EcoreUtil.convertToString(oldType, value)
				def newValue = EcoreUtil.createFromString(type, stringValue)
				instance.set(attribute, newValue)
			}
		}
	}
}
changeAttributeType = changeAttributeTypeImpl


// added 4/12/08 due to PCM case study
@description("In the metamodel, the super type of a class is replaced by one of its sub classes. In the model, nothing is modified.")
@label("Specialize Super Type")
specializeSuperTypeImpl = {
	@description("The class whose super type is specialized") EClass eClass,
	@description("The super type which is replaced") EClass toReplace,
	@description("The sub class by which is replaced") EClass replaceBy ->

	// constraints
	assert eClass.eSuperTypes.contains(toReplace) : "The super type to be replaced must be a super type of the class"
	assert replaceBy.eAllSuperTypes.contains(toReplace) : "The replacing super type must be a sub type of the replaced super type"
	
	// metamodel adaptation
	eClass.eSuperTypes.remove(toReplace)
	eClass.eSuperTypes.add(replaceBy)
}
specializeSuperType = specializeSuperTypeImpl


// added 4/12/08 due to PCM case study
@description("In the metamodel, a super type is removed from a class that is already inherited from another super class. In the model, nothing is changed, as this super type is superfluous.")
@label("Remove Superfluous Super Type")
removeSuperfluousSuperTypeImpl = {
	@description("The class from which the super type is removed") EClass eClass,
	@description("The super type to be removed") EClass superType ->

	// variables
	def List<EClass> superTypes = eClass.eSuperTypes
	
	// constraints
	assert superTypes.contains(superType) : "The super type to be removed actually has to be a super type"
	assert superTypes.any{s -> s.eAllSuperTypes.contains(superType)} : "The super type to be removed must be subsumed by one of the other super types"
	
	// metamodel adaptation
	eClass.eSuperTypes.remove(superType)
}
removeSuperfluousSuperType = removeSuperfluousSuperTypeImpl


// added 4/12/08 due to PCM case study
@description("In the metamodel, a super type is removed from a class. In the model, the values of the features inherited from that super type (including its super types) are deleted.")
@label("Remove Super Type")
removeSuperTypeImpl = {
	@description("The class from which the super type is removed") EClass eClass,
	@description("The super type to be removed") EClass superType = eClass.eSuperTypes[0] ->

	// constraints
	assert eClass.eSuperTypes.contains(superType) : "The super type to be removed actually has to be a super type of the class"
	
	// metamodel adaptation
	eClass.eSuperTypes.remove(superType)
	
	// model migration
	for(instance in eClass.allInstances) {
		for(feature in superType.eAllStructuralFeatures) {
			deleteFeatureValue(instance, feature)
		}
	}
}
removeSuperType = removeSuperTypeImpl


// added 4/12/08 due to PCM case study
@description("In the metamodel, the super type of a class is replaced by its super types. In the model, the values of the features that the class inherits from that super type (excluding its super types) are deleted.")
@label("Generalize Super Type")
generalizeSuperTypeImpl = {
	@description("The class of which the super type is replaced") EClass eClass,
	@description("The super type to be replaced by its super types") EClass superType = eClass.eSuperTypes[0] ->

	// constraints
	assert eClass.eSuperTypes.contains(superType) : "The super type to remove actually has to be a super type"
	
	// metamodel adaptation
	eClass.eSuperTypes.remove(superType)
	eClass.eSuperTypes.addAll(superType.eSuperTypes)
	
	// model migration
	for(instance in eClass.allInstances) {
		for(EStructuralFeature feature in superType.eStructuralFeatures) {
			deleteFeatureValue(instance, feature)
		}
	}
}
generalizeSuperType = generalizeSuperTypeImpl


// added 4/12/08
@description("In the metamodel, an element is renamed. In the model, nothing is changed.")
@label("Rename")
renameImpl = {
	@description("The metamodel element to be renamed") ENamedElement element,
	@description("The new name") String name = element.name ->
	
	// constraints
	assert element.eContainer() == null || element.eContainer().eContents().findAll{e -> e instanceof ENamedElement}.every{n -> !n.name.equals(name)} : "The name must not be already defined by the children of the element's parent."
	
	// metamodel adaptation
	element.name = name
	if(element instanceof EEnumLiteral) {
		element.literal = name
	}
}
rename = renameImpl


// added due to GMF case study
@description("In the metamodel, a number of features are combined in to a single feature by moving it over references to the same class. In the model, the values of the features are moved accordingly.")
@label("Combine Features over References")
combineFeatureImpl = {
	@description("The features to be combined") List<EStructuralFeature> features,
	@description("The references over which the features are moved (in the same order)") List<EReference> references ->

	def EClass eClass = references[0].eType
	def EStructuralFeature mainFeature = features[0]
	
	// constraints
	assert references.every{it -> it.eType == eClass} : "All references must have the same class as type"
	assert features.size() == references.size() : "There must be an equal number of features and references"
	assert references.every{it -> it.eContainingClass == features[references.indexOf(it)].eContainingClass} : "Each feature has to belong to its reference's class"
	
	// metamodel adaptation
	eClass.eStructuralFeatures.add(mainFeature)
	for(feature in features) {
		if(feature != mainFeature) {
			feature.delete()
		}
	}
	
	// model migration
	for(int i = 0; i < references.size(); i++) {
		def reference = references[i]
		def feature = features[i]
		for(instance in reference.eContainingClass.allInstances) {
			def value = instance.unset(feature)
			def ref = instance.get(reference)
			if(ref != null) {
				ref.set(mainFeature, value)
			}
		}
	}
}
combineFeature = combineFeatureImpl


//added due to GMF case study
@description("In the metamodel, a feature is propagated opposite to a number of references. More specifically, the feature is created in each of the classes which are sources of the references. In the model, the values of that feature are moved accordingly.")
@label("Propagate Feature over References")
propagateFeatureImpl = {
	@description("The feature to be propagated") EStructuralFeature mainFeature,
	@description("The references opposite to which the feature is propagated") List<EReference> references ->

	// variables
	def EClass eClass = mainFeature.eContainingClass
	
	// constraints
	assert references.every{it -> it.eType == eClass} : "Every reference has to target the class with the feature"
	
	// metamodel adaptation
	boolean first = true
	def features = []
	for(reference in references) {
		def feature = null
		if(first) {
			feature = mainFeature
			first = false
		}
		else {
			feature = mainFeature.clone()
		}
		features.add(feature)
		reference.eContainingClass.eStructuralFeatures.add(feature)
	}
	
	// model migration
	for(int i = 0; i < references.size(); i++) {
		def reference = references[i]
		def feature = features[i]
		for(instance in reference.eContainingClass.allInstances) {
			def ref = instance.get(reference)
			if(ref != null) {
				def value = ref.unset(mainFeature)
				instance.set(feature, value)
			}
		}
	}
}
propagateFeature = propagateFeatureImpl


// added due to GMF case study
@label("Replace Enumeration")
@description("In the metamodel, an enumeration is replaced by another one. More specifically, the enumeration is deleted and the other enumeration used instead. In the model, the values of this enumeration are replaced based on a mapping of literals.")
replaceEnumImpl = {
	@description("The enumeration to be replaced") EEnum toReplace,
	@description("The enumeration by which it is replaced") EEnum replaceBy,
	@description("The literals to be replaced") List<EEnumLiteral> literalsToReplace,
	@description("The literals by which they are replaced (in the same order)") List<EEnumLiteral> literalsReplaceBy ->

	// variables
	
	// constraints
	assert literalsToReplace.size() == literalsReplaceBy.size() : "The replacing and replaced literals must be of the same size"
	assert toReplace.eLiterals.containsAll(literalsToReplace) : "The replaced literals must belong to the replaced enumeration"
	assert replaceBy.eLiterals.containsAll(literalsReplaceBy) : "The replacing literals must belong to the replacing enumeration"
	
	// metamodel adaptation
	def attributes = toReplace.getInverse(emf.EAttribute.eType)
	for(attribute in attributes) {
		attribute.eType = replaceBy
	}
	toReplace.delete()
	
	// model migration
	for(attribute in attributes) {
		def eClass = attribute.eContainingClass
		for(instance in eClass.allInstances) {
			if(instance.isSet(attribute)) {
				def value = instance.get(attribute)
				int index = literalsToReplace.indexOf(value)
				instance.set(literalsReplaceBy[index])
			}
		}
	}
}
replaceEnum = replaceEnumImpl


// added 6/12/08
@description("In the metamodel, a class that is no longer used is deleted. In the model, nothing is changed.")
@label("Delete Class")
deleteClassImpl = {
	@description("The class to be deleted") EClass eClass ->

	// constraints
	assert eClass.getInverse(emf.ETypedElement.eType).isEmpty() : "The class must not be the target of a reference"
	assert eClass.eSubTypes.isEmpty() : "The class must not have sub classes"
	assert eClass.eSuperTypes.isEmpty() : "The class must not have super classes"
	
	// metamodel adaptation
	eClass.delete()
}
deleteClass = deleteClassImpl


// added 7/12/08 due to PCM case study
@description("In the metamodel, the type of a containment reference is specialized by a new sub class. In the model, the values of this reference are migrated to the new type.")
@label("Specialize Composition")
specializeCompositionImpl = {
	@description("The containment reference to be specialized") EReference reference,
	@description("The package in which the sub class is created") EPackage ePackage,
	@description("The name of the sub class") String name ->

	def EClass superType = reference.eType
	def EClass eClass = reference.eContainingClass
	
	// constraints
	assert reference.containment : "The reference has to be a containment reference"
	
	// metamodel adaptation
	def newType = ePackage.newEClass(name, [superType], false)
	reference.eType = newType
	
	// model migration
	for(instance in eClass.allInstances) {
		if(instance.isSet(reference)) {
			def value = instance.get(reference)
			if(reference.many) {
				for(v in value) {
					if(v.getEClass() == superType) {
						v.migrate(newType)
					}
					else {
						v.delete()
					}
				}
			}
			else if(value != null) {
				if(value.getEClass() == superType) {
					value.migrate(newType)
				}
				else {
					value.delete()
				}
			}
		}
	}
}
specializeComposition = specializeCompositionImpl


// added 7/12/08 due to PCM case study
@description("In the metamodel, a class is deleted. In the model, its instances are migrated to a class sharing the same super class.")
@label("Replace Class by Sibling")
@deprecated
replaceClassBySiblingImpl = {
	@description("The class to be replaced") EClass toReplace,
	@description("The class by which it is replaced") EClass replaceBy ->
	
	// constraints
	assert toReplace.eSuperTypes.size() == 1 : "The replaced class must have exactly one super class"
	assert replaceBy.eSuperTypes.size() == 1 : "The replacing class must have exactly one super class"
	assert toReplace.eSuperTypes[0] == replaceBy.eSuperTypes[0] : "The super classes of replaced and replacing class must be the same"
	assert toReplace.eSubTypes.isEmpty() : "The replaced class must not have sub classes"
	assert toReplace.eStructuralFeatures.isEmpty() : "The replaced class must not have any features"
	
	// metamodel adaptation
	for(reference in toReplace.getInverse(emf.EReference.eType)) {
		reference.eType = replaceBy
	}
	toReplace.delete()
	
	// model migration
	for(instance in toReplace.allInstances) {
		instance.migrate(replaceBy)
	}
}
replaceClassBySibling = replaceClassBySiblingImpl


@description("In the metamodel, a classifier is moved to a different package. In the model, nothing is changed.")
@label("Move Classifier")
moveClassifierImpl = {
	@description("The classifier to be moved") EClassifier classifier,
	@description("The package to which the classifier is moved") EPackage targetPackage ->

	// constraints
	assert classifier.ePackage != targetPackage : "The classifier must not be already part of the target package" 
	assert targetPackage.getEClassifier(classifier.name) == null : "A classifier with the same name exists in the target package"

	// metamodel adaptation
	targetPackage.eClassifiers.add(classifier)
}
moveClassifier = moveClassifierImpl


// added 16/12/2008 due to unicase case study
@description("In the metamodel, the sub class is deleted. In the model, all instances of this sub class are migrated to its super class.")
@label("Inline Sub Class")
inlineSubClassImpl = {
	@description("The class to be inlined") EClass subClass ->

	assert subClass.eSuperTypes.size() == 1 : "The sub class must have exactly one super type"
	assert subClass.eSubTypes.isEmpty() : "The sub class must not have sub types"
	assert subClass.eStructuralFeatures.isEmpty() : "The sub class must not have features"
	
	def superClass = subClass.eSuperTypes[0]
	
	// metamodel adaptation
	subClass.delete()
	
	// model migration
	for(instance in subClass.allInstances) {
		instance.migrate(superClass)
	}
}
inlineSubClass = inlineSubClassImpl


// added 21/12/2008 due to PCM case study
@description("In the metamodel, a number of feature are extracted to an existing class. More specifically, a containment reference to the extracted class is created and the features are replaced by features of the extracted class. In the model, the values of the features are moved accordingly to a new instance of the extracted class.")
@label("Extract Existing Class")
extractExistingClassImpl = {
	@description("The features to be extracted") List<EStructuralFeature> toReplace,
	@description("The extracted class") EClass extractedClass,
	@description("The features of the extracted class by which they are replaced (in the same order)") List<EStructuralFeature> replaceBy,
	@description("The name of the containment reference") String referenceName ->

	def EClass contextClass = toReplace[0].eContainingClass
	
	assert toReplace.size() == replaceBy.size() : "The replaced and replacing features must be of the same size"
	assert extractedClass.eAllStructuralFeatures.containsAll(replaceBy) : "The features to replace must be defined in the extracted class"
	assert toReplace.every{it.eType == replaceBy[toReplace.indexOf(it)].eType} : "The features must be of the same type"
	assert toReplace.every{it.many == replaceBy[toReplace.indexOf(it)].many} : "The features must be of the same multiplicity"
	
	def reference = contextClass.newEReference(referenceName, extractedClass, 1, 1, true)
	for(feature in toReplace) {
		feature.delete()
	}
	
	for(contextInstance in contextClass.allInstances) {
		def extractedInstance = extractedClass.newInstance()
		contextInstance.set(reference, extractedInstance)
		for(int i = 0; i < toReplace.size(); i++) {
			extractedInstance.set(replaceBy[i], contextInstance.unset(toReplace[i]))
		}
	}
}
extractExistingClass = extractExistingClassImpl


subtractSuperTypes = { EClass minuend, EClass subtrahend ->

	def superTypes = []
	
	superTypes.addAll(minuend.eAllSuperTypes)
	superTypes.add(0, minuend)
	
	superTypes.removeAll(subtrahend.eAllSuperTypes)
	superTypes.remove(subtrahend)
	
	return superTypes
}

subtractFeatures = { EClass minuend, EClass subtrahend ->

	def features = []
	def superTypes = subtractSuperTypes(minuend, subtrahend)
	
	for(superType in superTypes) {
		features.addAll(superType.eStructuralFeatures)
	}
	
	return features
}

// added 22/12/2008 due to GMF case study
@description("In the metamodel, a class is deleted. In the model, instances of this class are migrated to another class based on a mapping of features.")
@label("Replace Class")
replaceClassImpl = {
	@description("The class to be replaced") EClass toReplace,
	@description("The class by which it is replaced") EClass replaceBy,
	@description("The features to be replaced") List<EStructuralFeature> featuresToReplace = [],
	@description("The features by which they are replaced (in the same order)") List<EStructuralFeature> featuresReplaceBy = [] ->

	// constraints
	assert toReplace.getInverse(emf.EClass._eSuperTypes).isEmpty() : "The class to be replaced must not have sub types"
	assert featuresToReplace.size() == featuresReplaceBy.size() : "The replaced and replacing features have to be of the same size"
	assert toReplace.eAllStructuralFeatures.containsAll(featuresToReplace) : "The replace features must be defined in the replaced class"
	assert replaceBy.eAllStructuralFeatures.containsAll(featuresReplaceBy) : "The replacing features must be defined in the replacing class"
	assert subtractFeatures(toReplace, replaceBy).every{feature -> featuresToReplace.contains(feature)} : "The replace features must cover all features from the difference between the class to replace and the class by which it is replaced"
	
	// metamodel adaptation
	for(reference in toReplace.getInverse(emf.EReference.eType)) {
		reference.eType = replaceBy
	}
	toReplace.delete()
	
	// model migration
	for(instance in toReplace.allInstances) {
		instance.migrate(replaceBy)		
		for(int i = 0; i < featuresToReplace.size(); i++) {
			instance.set(featuresReplaceBy[i], instance.unset(featuresToReplace[i]))
		}
	}
}
replaceClass = replaceClassImpl


//added 4/1/2009 due to gmfmap case study
@label("Extract Subclass")
@description("In the metamodel, a feature is extracted into a new subclass and the feature is made mandatory. In the model, all instances of the superclass that have the feature set are migrated to the new subclass.")
extractSubClassImpl = {
	@description("The superclass from which the feature is extracted") EClass superClass,
	@description("The feature to be extracted") EStructuralFeature feature,
	@description("The name of the new subclass") String className ->

	// constraints
	assert superClass.eStructuralFeatures.contains(feature) : "The feature has to belong to the super class"
	assert superClass.eSubTypes.isEmpty() : "The super class may not have a sub class"
	
	// metamodel adaptation
	def ePackage = superClass.ePackage
	def subClass = ePackage.newEClass(className, [superClass])
	subClass.eStructuralFeatures.add(feature)
	feature.lowerBound = 1
	
	// model migration
	for(instance in superClass.instances) {
		if(instance.isSet(feature)) {
			instance.migrate(subClass)
		}
	}
}
extractSubClass = extractSubClassImpl


//added 31/12/2008 due to gmfgraph case study
@label("Imitate Superclass")
@description("In the metamodel, a superclass is removed from a subclass, while all its features are copied into the subclass. In the model, values are changed accordingly.")
imitateSuperTypeImpl = {
	@description("The subclass") EClass subClass,
	@description("The superclass") EClass superClass = subClass.eSuperTypes[0] ->

	def List<EStructuralFeature> features = superClass.eStructuralFeatures
	
	// constraints
	assert subClass.eSuperTypes.contains(superClass) : "The super class has to be a super type of the sub class"
	assert superClass.getInverse(emf.ETypedElement.eType).isEmpty() : "The super class must not be target of a reference"
	
	// metamodel adaptation
	subClass.eSuperTypes.remove(superClass)
	subClass.eSuperTypes.addAll(superClass.eSuperTypes)
	
	def clones = []
	for(feature in features) {
		def clone = feature.clone()
		subClass.eStructuralFeatures.add(clone)
		clones.add(clone)
	}
	
	// model migration
	for(instance in subClass.allInstances) {
		for(int i = 0; i < features.size(); i++) {
			instance.set(clones[i], instance.unset(features[i]))
		}
	}
}
imitateSuperType = imitateSuperTypeImpl


//added 19/02/2009 due to GMF case study
@label("Extract and Group Attribute")
@description("In the metamodel, an attribute is extracted into a new class. This extracted class is contained by an existing container class and referenced from the context class. In the model, an instance of the extracted class is created for each different value of the extracted attribute.")
extractAndGroupAttributeImpl = {
	@description("The attribute to be extracted") EAttribute extractedAttribute,
	@description("The package in which the extracted class is created") EPackage contextPackage,
	@description("The name of the extracted class") String extractedClassName,
	@description("The reference from the context class to the extracted class") String referenceName,
	@description("The container class for the extracted class") EClass containerClass,
	@description("The name of the containment reference from the container class to the extracted class") String containerReferenceName ->

	def EClass contextClass = extractedAttribute.eContainingClass
	
	// constraints
	assert !extractedAttribute.many : "The extracted attribute must be single-valued"

	// metamodel adaptation
	def extractedClass = contextPackage.newEClass(extractedClassName)
	extractedClass.eStructuralFeatures.add(extractedAttribute)
	extractedAttribute.lowerBound = 1
	
	def reference = contextClass.newEReference(referenceName, extractedClass, 0, 1, false)
	
	def containerReference = containerClass.newEReference(containerReferenceName, extractedClass, 0, -1, true)
	
	// model migration
	for(contextElement in contextClass.allInstances) {
		def value = contextElement.unset(extractedAttribute)
		if(value != null) {
			def containerElement = contextElement;
			while(containerElement != null && !(containerElement.instanceOf(containerClass))) {
				containerElement = containerElement.getContainer()
			}
			if(containerElement != null) {
				def extractedElement = containerElement.get(containerReference).find{e -> value.equals(e.get(extractedAttribute))}
				if(extractedElement == null) {
					extractedElement = extractedClass.newInstance()
					extractedElement.set(extractedAttribute, value)
					containerElement.get(containerReference).add(extractedElement)
				}
				contextElement.set(reference, extractedElement)
			}
		}
	}
}
extractAndGroupAttribute = extractAndGroupAttributeImpl


//added 19/02/2009 due to GMF case study
@label("Copy Feature")
@description("In the metamodel, a feature is copied, giving it a new name. In the model, the values are copied, accordingly.")
copyFeatureImpl = {
	@description("The feature to be copied") EStructuralFeature feature,
	@description("The name of the copy") String name ->

	def EClass contextClass = feature.eContainingClass 
	
	// metamodel adaptation
	def EStructuralFeature copiedFeature = feature.clone()
	copiedFeature.name = name
	contextClass.eStructuralFeatures.add(copiedFeature)
	if(copiedFeature instanceof EReference && copiedFeature.containment) {
		copiedFeature.containment = false
	}
	
	// model migration
	for(instance in contextClass.allInstances) {
		instance.set(copiedFeature, instance.get(feature))
	}
}
copyFeature = copyFeatureImpl


//added 19/02/2009 due to GMF case study
@label("Flatten Hierarchy")
@description("In the metamodel, a containment hierarchy is flattened. More specifically, the reference to denote the root as well as the reference to denote the children are replaced by a containment reference. In the model, the corresponding hierarchies are flattened accordingly.")
flattenHierarchyImpl = {
	@description("The reference to denote the root node") EReference rootReference,
	@description("The reference to denote the children nodes") EReference childrenReference,
	@description("The reference which replaces the containment hierarchy") String referenceName ->

	def EClass rootClass = rootReference.eContainingClass
	def EClass nodeClass = rootReference.eType
	
	// constraints
	assert !rootReference.many && rootReference.containment : "The root reference must be a single-valued containment reference."
	assert childrenReference.many && childrenReference.containment : "The children reference must be a multi-valued containment reference."
	assert nodeClass.eStructuralFeatures.contains(childrenReference) : "The children reference must be defined by the node class."
	assert childrenReference.eType == nodeClass : "The type of the children reference must be the node class."
	
	// metamodel adaptation
	rootReference.delete()
	childrenReference.delete()
	def containerReference = rootClass.newEReference(referenceName, nodeClass, 0, -1, true)
	
	// model migration
	visitNode = { root, node ->
		def children = node.unset(childrenReference)
		root.get(containerReference).addAll(children)
		for(child in children) {
			visitNode(root, child)
		}
	}
	
	for(root in rootClass.allInstances) {
		def node = root.unset(rootReference)
		if(node != null) {
			root.get(containerReference).add(node)
			visitNode(root, node)
		}
	}
}
flattenHierarchy = flattenHierarchyImpl


@label("Make Reference Containment")
@description("In the metamodel, a reference is made containment. In the model, its values are replaced by copies.")
makeContainmentImpl = {
	@description("The reference") EReference reference ->

	def EClass contextClass = reference.eContainingClass
	
	// constraints
	assert !reference.containment : "The reference must not already be containment."

	// metamodel adaptation
	reference.containment = true
	
	// model migration
	for(instance in contextClass.allInstances) {
		if(reference.many) {
			def values = instance.unset(reference)
			for(value in values) {
				instance.get(reference).add(value.clone())
			}
		}
		else {
			def value = instance.unset(reference)
			if(value != null) {
				instance.set(reference, value.clone())
			}
		}
	}
}
makeContainment = makeContainmentImpl


//added 27/2/09 due to GMF case study
@label("Replace Literal")
@description("In the metamodel, an enum literal is removed and replaced by another one. In the model, the enum's values are replaced accordingly.")
replaceLiteralImpl = {
	@description("The enum literal to replace") EEnumLiteral toReplace,
	@description("The enum literal by which it is replaced") EEnumLiteral replaceBy ->

	def EEnum contextEnum = toReplace.eEnum

	// constraints
	assert contextEnum.eLiterals.contains(replaceBy) : "The enum literal by which it is replace must belong to the same enum."
	
	// metamodel adaptation
	toReplace.delete()
	
	// model migration
	def attributes = contextEnum.getInverse(emf.EAttribute.eType)
	for(attribute in attributes) {
		def contextClass = attribute.eContainingClass
		for(instance in contextClass.allInstances) {
			if(instance.get(attribute) == toReplace) {
				instance.set(attribute, replaceBy)
			}
		}
	}
}
replaceLiteral = replaceLiteralImpl


@label("Delete Package")
@description("In the metamodel, an empty package is deleted.")
deletePackageImpl = {
	@description("The package to be deleted") EPackage ePackage ->
	
	assert ePackage.eClassifiers.isEmpty() : "The package must not contain classifiers"
	assert ePackage.eSubpackages.isEmpty() : "The package must not contain subpackages"
	
	ePackage.delete()
}
deletePackage = deletePackageImpl


//added 18/1/2009 due to gmfgen case study
@label("Make Class Abstract")
@description("In the metamodel, a class is made abstract. In a model, instances of this class are migrated to a chosen subclass.")
makeAbstractImpl = {
	@description("The class to be made abstract") EClass eClass,
	@description("The subclass to which instances are migrated") EClass subClass ->

	// constraints
	assert !eClass.'abstract' : "The class is not yet abstract"
	assert subClass.eAllSuperTypes.contains(eClass) : "The class has to be a super type of the sub class"
	
	// metamodel adaptation
	eClass.'abstract' = true
	
	// model migration
	for(instance in eClass.instances) {
		instance.migrate(subClass)
	}
}
makeAbstract = makeAbstractImpl


//added 31/12/2008 due to gmfgraph case study
@label("Generalize Attribute")
@description("In the metamodel, the multiplicity of an attribute is generalized. In the model, nothing is changed.")
generalizeAttributeImpl = {
	@description("The attribute to be generalized") EAttribute attribute,
	@description("The new lower bound of the attribute") int lowerBound = attribute.lowerBound,
	@description("The new upper bound of the attribute") int upperBound = attribute.upperBound ->
	
	// constraints
	assert lowerBound <= attribute.lowerBound && (upperBound >= attribute.upperBound || upperBound == -1) : "The multiplicity must be the same or more general"
	
	// metamodel adaptation
	attribute.lowerBound = lowerBound
	attribute.upperBound = upperBound
}
generalizeAttribute = generalizeAttributeImpl


//added 31/12/2008 due to gmfgraph case study
@label("Specialize Reference Type")
@description("In the metamodel, the type of a reference can be specialized to its subclass, in case it is abstract and has only one subclass. In the model, nothing is changed.")
specializeReferenceTypeImpl = {
	@description("The reference whose type is specialized") EReference reference,
	@description("The new type of the reference") EClass type ->
	
	def EClass oldType = reference.eType

	// constraints
	assert oldType.'abstract' || oldType.'interface' : "The old type of the reference must be abstract"
	assert type.eSuperTypes.contains(oldType) : "The new type of the reference must be a subclass of its old type"
	assert oldType.eSubTypes.size() == 1 : "The old type must not have any other subclass"

	// metamodel adaptation
	reference.eType = type
}
specializeReferenceType = specializeReferenceTypeImpl


//added 1/5/2009 due to the GMF case study
@label("New Opposite Reference")
@description("In the metamodel, an opposite is created for a reference. In the model, the opposite direction needs to be set.")
newOppositeReferenceImpl = {
	@description("The reference for which the opposite is created") EReference reference,
	@description("The name of the opposite reference") String name,
	@description("The lower bound of the opposite reference") int lowerBound = 0,
	@description("The upper bound of the opposite reference") int upperBound = reference.containment ? 1 : -1,
	@description("Whether the opposite reference is changeable") boolean changeable = true ->

	// variables
	def EClass eClass = reference.eType
	def EClass type = reference.eContainingClass
	
	// constraints
	assert reference.eOpposite == null : "The reference must not already have an opposite"
	assert !reference.containment || upperBound == 1 : "In case of a containment reference, the upper bound of the opposite reference must be 1."
	assert reference.containment || upperBound == -1: "In case of a cross reference, the upper bound of the opposite reference must be -1."

	// metamodel adaptation
	def opposite = eClass.newEReference(name, type, lowerBound, upperBound, false, reference)
	opposite.changeable = changeable
}
newOppositeReference = newOppositeReferenceImpl


//added 1/5/2009 due to the GMF case study
@label("Drop Opposite Relationship")
@description("In the metamodel, the opposite relationship between to references is dropped. In the model, nothing needs to be done.")
dropOppositeImpl = {
	@description("The reference whose opposite relationship should be dropped") EReference reference ->

	// constraints
	assert reference.eOpposite != null : "Reference must have an opposite"
	
	// metamodel adaptation
	reference.eOpposite = null
}
dropOpposite = dropOppositeImpl


//added 1/5/2009 due to the GMF case study
@label("Make Feature Volatile")
@description("In the metamodel, a feature is made volatile. In the model, its values have to be deleted.")
makeFeatureVolatileImpl = {
	@description("The feature to be made volatile") EStructuralFeature feature,
	@description("Whether the feature is transient") boolean trans = true,
	@description("Whether the feature is derived") boolean derived = true,
	@description("Whether the feature is changeable") boolean changeable = false ->

	// constraints
	assert !feature.'volatile' : "Feature must not be volatile"
	
	// metamodel adaptation
	feature.'volatile' = true
	feature.'transient' = trans
	feature.derived = derived
	feature.changeable = changeable
	if(feature instanceof EReference && feature.eOpposite != null) {
		dropOpposite(feature)
	}

	// model migration
	for(instance in feature.eContainingClass.allInstances) {
		deleteFeatureValue(instance, feature)
	}
}
makeFeatureVolatile = makeFeatureVolatileImpl


//added 1/5/2009 due to the GMF case study
@label("Volatile to Opposite Reference")
@description("In the metamodel, a reference is changed from being volatile to an opposite. In the model, the opposite direction needs to be set.")
volatileToOppositeImpl = {
	@description("The reference which is changed from volatile to opposite") EReference reference,
	@description("The reference which is going to be the opposite") EReference opposite,
	@description("Whether the reference is going to be changeable") boolean changeable = true ->

	// constraints
	assert reference.eOpposite == null : "Reference must not already have an opposite"
	assert reference.'volatile' : "Reference must be volatile"
	assert reference.eType == opposite.eContainingClass && reference.eContainingClass == opposite.eType : "Reference and opposite must be compatible with each other"
	
	// metamodel adaptation
	reference.'volatile' = false
	reference.'transient' = false
	reference.derived = false
	reference.changeable = changeable
	opposite.eOpposite = reference
}
volatileToOpposite = volatileToOppositeImpl


//added 1/5/2009 due to the GMF case study
@label("Pull up Operation")
@description("In the metamodel, a number of operations are pulled up into a common super class. In the model, nothing needs to be done.")
pullOperationImpl = {
	@description("The operations to be pulled up") List<EOperation> operations,
	@description("The super class to which the operations are pulled") EClass targetClass = operations[0].eContainingClass.eSuperTypes[0] ->

	// variables
	def EOperation mainOperation = operations[0]

	// constraints
	assert operations.every{operation -> mainOperation.eType == operation.eType} : "The operations' types have to be the same"
	assert operations.every{operation -> mainOperation.lowerBound == operation.lowerBound && mainOperation.upperBound == operation.upperBound} : "The operations' multiplicities have to be the same"
	assert operations.every{operation -> operation.eContainingClass.eSuperTypes.contains(targetClass)} : "The operations' classes must have a common super type"
	
	// metamodel adaptation
	targetClass.eOperations.add(mainOperation)
	for(operation in operations) {
		if(operation != mainOperation) {
			operation.delete()
		}
	}	
}
pullOperation = pullOperationImpl


//added 1/5/2009 due to the GMF case study
@label("Push down Operation")
@description("In the metamodel, an operation is pushed down to its sub classes. In the model, nothing needs to be done.")
pushOperationImpl = {
	@description("The operation to be pushed down") EOperation operation ->
	
	// variables
	EClass superClass = operation.eContainingClass
	List<EClass> subClasses = superClass.eSubTypes
	
	// metamodel adaptation
	def first = true
	for(subClass in subClasses) {
		if(first) {
			subClass.eOperations.add(operation)
		}
		else {
			def clone = operation.clone()
			subClass.eOperations.add(clone)
		}
		first = false;
	}
}
pushOperation = pushOperationImpl


//added 1/5/2009 due to the GMF case study
@label("Operation to Volatile Feature")
@description("In the metamodel, an operation is transformed into a volatile feature. In the model, nothing needs to be done.")
operationToVolatileImpl = {
	@description("The operation to be transformed") EOperation operation ->

	// constraints
	assert operation.eParameters.isEmpty() : "The operation must not have parameters"
	
	// metamodel adaptation
	def EClass eClass = operation.eContainingClass
	def EClassifier type = operation.eType
		
	def name = operation.name
	if(name.startsWith("get")) {
		name = name.substring(3)
		name = name.substring(0,1).toLowerCase() + name.substring(1)
	}
	
	def EStructuralFeature feature = null 
	if(type instanceof EClass) {
		feature = eClass.newEReference(name, type, operation.lowerBound, operation.upperBound)
	}
	else {
		feature = eClass.newEAttribute(name, type, operation.lowerBound, operation.upperBound)
	}
	feature.'volatile' = true
	feature.'transient' = true
	feature.derived = true
	feature.changeable = false
	
	operation.delete()
}
operationToVolatile = operationToVolatileImpl


@label("Document Metamodel Element")
@description("In the metamodel, a metamodel element is documented. Nothing is changed in the model.")
documentImpl = {
	@description("The metamodel element to be documented") EModelElement element,
	@description("The comment for documentation") String documentation = entry.value ->

	EAnnotation annotation = element.getEAnnotation("http://www.eclipse.org/emf/2002/GenModel")
	EObject entry = annotation != null ? annotation.details.find{d -> d.key == "documentation"} : null
	
	if(annotation == null) {
		annotation = element.newEAnnotation("http://www.eclipse.org/emf/2002/GenModel")
	}
	if(entry == null) {
		entry = annotation.newEStringToStringMapEntry()
		entry.key = "documentation"
	}
	entry.value = documentation
}
document = documentImpl


@label("New GMF Constraint")
@description("In the metamodel, a new constraint is introduced. Nothing is changed in the model.")
newGMFConstraintImpl = {
	@description("The metamodel element in which context the constraint is created") EModelElement element,
	@description("The OCL expression of the constraint") String ocl,
	@description("The description of the constraint") String description ->

	def annotation = element.newEAnnotation("http://www.eclipse.org/gmf/2005/constraints")
	annotation.newEStringToStringMapEntry("ocl", ocl)
	annotation.newEStringToStringMapEntry("description", description)
}
newGMFConstraint = newGMFConstraintImpl


@label("Not Changeable to Suppressed Set Visibility")
@description("In the metamodel, a reference is made changeable, and at the same time its setter is suppressed. Nothing is changed in the model.")
notChangeableToSuppressedSetVisibilityImpl = {
	@description("The reference to be made changeable") EReference reference ->

	reference.changeable = true
	def annotation = reference.newEAnnotation("http://www.eclipse.org/emf/2002/GenModel")
	annotation.newEStringToStringMapEntry("suppressedSetVisibility", "true");
}
notChangeableToSuppressedSetVisibility = notChangeableToSuppressedSetVisibilityImpl


@label("Suppressed Set Visibility to Not Changeable")
@description("In the metamodel, the setter of a reference is made visible again, and at the same time it is made non-changeable. Nothing is changed in the model.")
suppressedSetVisibilityToNotChangeableImpl = { 
	@description("The reference whose setter is made visible again") EReference reference ->

	EAnnotation annotation = reference.eAnnotations.find{a -> a.details.find{d -> d.key == "suppressedSetVisibility"} != null}
	EObject entry = annotation.details.find{d -> d.key == "suppressedSetVisibility"}
	
	assert entry != null : "Suppressed Set Visibility must be present"
	
	reference.changeable = false
	if(annotation.details.size() > 1) {
		entry.delete()
	}
	else {
		annotation.delete()
	}
}
suppressedSetVisibilityToNotChangeable = suppressedSetVisibilityToNotChangeableImpl


@label("Sub Classes to Enumeration")
@description("In the metamodel, the subclasses of a class are replaced by an enumeration. " +
		"An enumeration with literals for all subclasses is created and an enumeration attribute is created in the class. " +
		"Finally, all subclasses are deleted, and the class is made concrete. " +
		"In the model, instances of a subclass are migrated to the class, setting the enumeration attribute to the appropriate literal.")
subClassesToEnumerationImpl = {
	@description("The context class") EClass contextClass,
	@description("The name of the enumeration attribute") String attributeName,
	@description("The package in which the enumeration is created") EPackage ePackage = contextClass.ePackage,
	@description("The name of the enumeration") String enumName ->
	
	// constraints
	assert !isConcrete(contextClass) : "The context class must be abstract"
	assert !contextClass.eSubTypes.isEmpty() : "The context class must have sub types"
	assert contextClass.eSubTypes.every{c -> c.eSubTypes.isEmpty()} : "The sub types must not have sub types again"

	// metamodel adaptation
	def enumeration = ePackage.newEEnum(enumName)
	def attribute = contextClass.newEAttribute(attributeName, enumeration, 1, 1)

	contextClass.'abstract' = false
	
	def subClasses = []
	int i = 0
	for(EClass subClass in contextClass.eSubTypes) {
		def literal = enumeration.newEEnumLiteral(subClass.name)
		literal.value = i
		subClasses.add(subClass)
		subClass.delete()
		i++
	}

	// model migration
	i = 0
	for(EClass subClass in subClasses) {
		def literal = enumeration.eLiterals[i]
		for(instance in subClass.instances) {
			instance.migrate(contextClass)
			instance.set(attribute, literal)
		}
		i++
	}
}
subClassesToEnumeration = subClassesToEnumerationImpl


@label("Enumeration to Sub Classes")
@description("In the metamodel, an enumeration attribute of a class is replaced by subclasses. " +
		"The class is made abstract, and a subclass is created for each literal of the enumeration. " +
		"The enumeration attribute is deleted and also the enumeration, if not used otherwise. " +
		"In the model, instances the class are migrated to the appropriate subclass according to the value of the enumeration attribute.")
enumerationToSubClassesImpl = {
	@description("The enumeration attribute") EAttribute enumAttribute,
	@description("The package in which the subclasses are created") EPackage ePackage = enumAttribute.eContainingClass.ePackage ->
	
	// variables
	EClass contextClass = enumAttribute.eContainingClass
	EClassifier enumeration = enumAttribute.eType
	
	// constraints
	assert isConcrete(contextClass) : "The context class must be concrete"
	assert contextClass.eSubTypes.isEmpty() : "The context class must not have sub types"
	assert enumeration instanceof EEnum : "The type of the attribute must be an enumeration"
	
	// metamodel adaptation
	def subClasses = []
	for(literal in enumeration.eLiterals) {
		def subClass = ePackage.newEClass(literal.name, [contextClass])
		subClasses.add(subClass)
	}

	enumAttribute.delete()
	if(enumeration.getInverse(emf.ETypedElement.eType).isEmpty()) {
		enumeration.delete()
	}
	
	contextClass.'abstract' = true
	
	// model migration
	for(instance in contextClass.instances) {
		def value = instance.unset(enumAttribute)
		int index = enumeration.eLiterals.indexOf(value)
		instance.migrate(subClasses[index])
	}
}
enumerationToSubClasses = enumerationToSubClassesImpl


@label("Partition Reference")
@description("In the metamodel, a reference is partitioned into a number of references according to its type. " +
		"A sub reference is created for each subclass of the reference's type. " +
		"Finally, the original reference is deleted. " +
		"In the model, the value of the reference is partitioned accordingly.")
partitionReferenceImpl = {
	@description("The reference to be partitioned") EReference reference ->
	
	EClass contextClass = reference.eContainingClass
	EClass type = reference.eReferenceType
	
	assert !isConcrete(type) : "The type of the reference must be abstract"
	assert reference.many : "The reference must be multi-valued"
	
	// metamodel adaptation
	def subReferences = []
	for(subClass in type.eSubTypes) {
		def name = subClass.name.substring(0,1).toLowerCase() + subClass.name.substring(1)
		def subReference = contextClass.newEReference(name, subClass, 0, -1, reference.containment)
		subReferences.add(subReference)
	}
	
	reference.delete()
	
	// model migration
	for(instance in contextClass.allInstances) {
		def values = instance.unset(reference)
		for(value in values) {
			def subReference = subReferences.find{r -> value.instanceOf(r.eReferenceType)}
			instance.get(subReference).add(value)
		}
	}
}
partitionReference = partitionReferenceImpl
